
;*========================================================================
;*
;* AESFAST Public Domain GEM bindings.
;*  
;* Maintenance:
;*  06/20/89 - v1.3: Added check for negative object index in obj_parent.
;*  08/28/89 - v1.3: Renamed objrb_which to obj_rbfind, and added logic
;*                   to find object in arbitrary state.
;*  05/26/90 - v1.4: Added obj_rbselect() function.
;*========================================================================
 
;*************************************************************************
;*
;* AESUTOB4.S - Object-related utilities 4 of n.
;*  Non-standard utility functions.
;*
;*************************************************************************
         
          .include  "gemfast.sh"        ; Pull in header file.

;-------------------------------------------------------------------------
; obj_rbfind  - Return index of selected radio button, -1 if none selected.
;         
;               More generally speaking, this returns the index of the first 
;               child of the specified parent which has one or more of the
;               same bits set as in the 'state' value passed to this function.
;
;  int obj_rbfind(tree, parentobj, state);
;
;|v1.4: This obj_rbselect requires that this routine must not modify a0.
;-------------------------------------------------------------------------

_obj_rbfind::
          .cargs    #4,.ptree.l,.obj,.state
          exg       a1,d3                       ; save d3 in a1
          move.l    .ptree(sp),a0
          move.w    .obj(sp),d2
          move.w    d2,d0
          muls      #OBJ_SIZ,d0
          move.w    ob_head(a0,d0.l),d0
          bmi.s     .notfound
.loop:
          move.w    d0,d1
          muls      #OBJ_SIZ,d1
          move.w    ob_state(a0,d1.l),d3
          and.w     .state(sp),d3
          bne.s     .done
          move.w    ob_next(a0,d1.l),d0
          cmp.w     d0,d2
          bne.s     .loop
.notfound:
          moveq.l   #-1,d0
.done:
          exg       a1,d3               ; restore d3 from a1
          tst.w     d0                  ; insure CCR return matches d0
          rts

;-------------------------------------------------------------------------
; obj_parent - Return the parent of a given object.
;              By definition, the root object has no parent, 0 is returned.
;              An unusual object tree could throw this routine into a loop,
;              but in a standard tree, all the next/tail pointers should
;              be in good shape, and a parent will be found.
;             
;  int obj_parent(tree, cur_object);
;
;|v1.4: This obj_rbselect requires that this routine must not modify a0.
;-------------------------------------------------------------------------

_obj_parent::
          .cargs    #4,.ptree.l,.obj.w
          
          move.l    .ptree(sp),a0
          move.w    .obj(sp),d0
          ble.s     noparent
obj_parent:                             ; v1.4: internal entry point
parent_loop:
          move.w    d0,d2
          muls      #OBJ_SIZ,d0
          move.w    ob_next(a0,d0.l),d1
          move.w    d1,d0
          muls      #OBJ_SIZ,d1
          cmp.w     ob_tail(a0,d1.l),d2
          bne.s     parent_loop
noparent:
          tst.w    d0
          rts
          
;-------------------------------------------------------------------------
; obj_rbselect - De-select current radio button, select a new one.
;   Returns the index of the previously-selected radio button.
;
;  int obj_rbselect(tree, selobjct);
;-------------------------------------------------------------------------

_obj_rbselect::
          .cargs    #4,.ptree.l,.selobj.w,.state.w
          
          move.w    .selobj(sp),d0
          ble.s     .done
          move.l    .ptree(sp),a0
          bsr       obj_parent
          
          move.w    .state(sp),-(sp)
          move.w    d0,-(sp)
          move.l    a0,-(sp)
          bsr       _obj_rbfind
          addq.l    #8,sp
          tst.w     d0
          bmi.s     .done
          
          move.w    d0,d2
          muls      #OBJ_SIZ,d2
          move.w    .state(sp),d1
          not.w     d1
          and.w     d1,ob_state(a0,d2.w)

          move.w    .selobj(sp),d2
          muls      #OBJ_SIZ,d2
          not.w     d1
          or.w      d1,ob_state(a0,d2.w)
.done:
          tst.w     d0
          rts

